/* readword.c -- Copyright 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* readword -- read the next word of input
 * Liam Quin, August 1993 and later...
 *
 * $Id: readword.c,v 1.4 93/11/28 15:17:00 lee Exp $ 
 */

#include "error.h"
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef BSD
# include <strings.h>
#else
# include <string.h>
#endif

#include "globals.h" /* defines and declarations for database filenames */

#include "fileinfo.h"
#include "wordinfo.h"
#include "wordrules.h"
#include "filter.h"
#include "emalloc.h"
#include "addfile.h"
#include "Revision.h"

#ifdef SYSV
#define TOLOWER(ch) ch = tolower(ch)
#else
#define TOLOWER(ch) if (isupper(ch)) ch = tolower(ch)
#endif

extern int TooCommon();

/** System calls and library routines used in this file: **/
/** System calls: **/
/** Library Functions: **/
#ifndef tolower
 extern int tolower();
#endif

/** Functions within this file that need declaring: **/
static t_WordInfo * fReadWord(/* Stream or char ** */);
static void NewFile();
t_WordInfo *FReadWord(/*FileInfo*/);
t_WordInfo *SpReadWord(/* char **Stringpp, **Startp */);

/**/

extern int AsciiTrace; /* provide increasingly verbose info if not zero */

static int LastChar = 0;
static char *ThisWord = 0;
static char *ThatWord = 0;
static unsigned long BytesRead = 0;
static int StringMode = 0;

/* Actual character input -- dependant on StringMode */
#define GetChar(Stream) \
    ((StringMode) ? \
	(** (char **) (Stream) == '\0') ? EOF : (++BytesRead, *((*(char **) (Stream))++)) \
	: \
	(++BytesRead, getc(Stream)))

static char *WordStart = 0;

t_WordInfo *
SpReadWord(Stringpp, Startp, IgnoreCommon)
    char **Stringpp;
    char **Startp;
{
    t_WordInfo *Result;

    if (!Stringpp || !*Stringpp || !**Stringpp) {
	NewFile((t_FID) 0);
	return (t_WordInfo *) 0;
    }

    StringMode = 1;
    if (Startp) {
	WordStart = (*Stringpp);
    }
    Result = fReadWord((t_FileInfo *) Stringpp, IgnoreCommon);
    StringMode = 0;
    if (Startp) {
	*Startp = WordStart;
    }
    return Result;
}

/* use two static storage areas so we can be called twice in a row.
 * This is necessary to implement the WPF_LASTINBLOCK flag.
 */
static t_WordInfo This, That;
static int WordInBlock = -1;
static long LastPos = 0L;
static int SawCommon = 0;
static int SawLetters = 0;
static int SawPunctuation = 0;
static int BlockInFile = 0L;
static unsigned long LastBlock;

t_WordInfo *
FReadWord(FileInfo)
    t_FileInfo *FileInfo;
{
    static t_FID LastFid = ~(t_FID)0;

    if (FileInfo->FID != LastFid) {
	LastFid = FileInfo->FID;
	NewFile(LastFid);
    }

    /* 2nd argument to fReadWord is whether to ignore common words */
    return fReadWord(FileInfo->Stream, 1);
}

static void
NewFile(FID)
    t_FID FID;
{
    if (!ThisWord) {
	ThisWord = emalloc("ReadWord::ThisWord", MaxWordLength + 1);
	ThatWord = emalloc("ReadWord::ThatWord", MaxWordLength + 1);
    }

    This.WID = (t_WID) 0;
    This.WordPlaces = (t_WordPlace *) 0;
    This.DataBlock = (char *) 0;
    This.WordPlacesInHere = 0;

    This.WordPlace.FID = This.FID = FID;
    That = This; /* structure copy */

    /* the two structures differ in having pointers to two different
     * string buffers, so we have to restore the pointers:
     */

    This.Word = ThisWord;
    That.Word = ThatWord;

    WordInBlock = (-1); /* none, yet! */
    LastPos = BlockInFile = LastBlock = 0L;
    BytesRead = 0L;
    SawCommon = SawPunctuation = SawLetters = 0;
}

static t_WordInfo *
fReadWord(Stream, IgnoreCommon)
    FILE *Stream;
    int IgnoreCommon;
{
    static int ThisOrThat = 0;
    register char *q;
    t_WordInfo *WordInfo;
    char *Buffer;
    int ch;
    unsigned long Start;

    if (ThisOrThat) {
	WordInfo = &This;
	WordInfo->Word = ThisWord;
    } else {
	WordInfo = &That;
	WordInfo->Word = ThatWord;
    }

    /* The following loop is simply to avoid needless recursion. */
    for (;;) {
	int NumberOfValidCharactersInWord;

	q = Buffer = WordInfo->Word;

	/* Skip non-word characters */
	while ((ch = GetChar(Stream)) != EOF && !StartsWord(ch)) {

	    if (isdigit(ch)) {

		/* the character was something that is not the start of the
		 * word, but it might not have been a space -- for example,
		 * if we're not indexing numbers, consider 0xFFE6, which
		 * we don't want to index as xFFE6!
		 * Also, 3.14159 is only one `word'...
		 * On the other hand, tc|wyse300|xx| includes "wyse300"
		 * or "wyse" (depending on whether digits are allowed
		 * inside words), but not "tc|wyse300|xx|".
		 */

		Start = BytesRead - 1;
		if (StringMode) {
		    WordStart = (*(char ** )Stream) - 1;
		}

		do {
		    ch = GetChar(Stream);
		    if (ispunct(ch)) {
			SawPunctuation = 1;
		    }
		} while (
		    isdigit(ch) ||
		    ch=='.' ||
		    EndsWord(ch) ||
		    OnlyWithinWord(ch)
		);

		if (ch == EOF) break;

		if ((BytesRead - 1) - Start >= MinWordLength) {
		    BlockInFile = Start / FileBlockSize;

		    if (BlockInFile != LastBlock) {
			LastBlock = BlockInFile;
			WordInBlock = 0;
		    } else {
			++WordInBlock;
		    }
		    if (!IgnoreCommon) {
			LastPos = BytesRead - 1;
			WordInfo->Word = "123456";
			WordInfo->Length = 6;
			return WordInfo;
		    }
		}
	    } else if (ispunct(ch)) {
		SawPunctuation = 1;
	    }
	}

	/* ASSERT: we have read at least one character */

	if (ch == EOF) {
	    return (t_WordInfo *) 0;
	}

	Start = BytesRead - 1;

	/* remember the last character read: */

	if (isupper(ch)) {
	    WordInfo->WordPlace.Flags = WPF_UPPERCASE;
	    ch = tolower(ch);
	} else {
	    WordInfo->WordPlace.Flags = 0;
	}

	if (StringMode) {
	    WordStart = (*(char ** )Stream) - 1;
	}

	for (;;)  {

	    if (q - Buffer < MaxWordLength) {
		TOLOWER(ch);
		*q++ = LastChar = ch;
	    } else {
		LastChar = ch;
	    }

	    ch = GetChar(Stream);

	    if (EndsWord(ch)) {
		continue; /* OK, still inside the word */
	    } else if (OnlyWithinWord(ch)) {
		if (EndsWord(LastChar)) {
		    continue; /* OK, e.g. ch is ' in can't, t will be next */
		} else {
		    /* e.g. just seen 2nd ' in boy'' at end of quotation.
		     * Reject both this and the previous character:
		     */

		    /* reject previous character: */
		    if (q - Buffer <= MaxWordLength && q[-1] == LastChar) {
			*--q = '\0';
		    }

		    break;
		}
	    } else {
		if (OnlyWithinWord(LastChar)) {
		    /* reject the previous character: */
		    if (q - Buffer < MaxWordLength) {
			*--q = '\0';
		    }
		} else {
		    LastChar = 0;
		}
		break;
	    }
	}

	*q = '\0';

	/* following two lines split out to avoid GCC bug */
	WordInfo->Length = q - Buffer;
	if (WordInfo->Length < MinWordLength) {
	    register char *p;

	    *++q = ch;
	    *q = '\0';
	    for (p = Buffer; p < q; p++) {
		if (isalpha(*p)) {
		    SawLetters = 1;
		    if (SawPunctuation) break;
		} else if (ispunct(*p)) {
		    SawPunctuation = 1;
		    if (SawLetters) break;
		}
	    }

	    if (LastChar) {
		if (ispunct(LastChar)) {
		    SawPunctuation = 1;
		}
		LastChar = 0;
	    }

	    continue;
	}

	/* We have to do this now to get common words right,
	 * but we've left it as late as possible in order to hope to
	 * avoid doing the relatively expensive division:
	 */
	BlockInFile = Start / FileBlockSize;

	if (BlockInFile != LastBlock) {
	    LastBlock = BlockInFile;
	    WordInBlock = (-1);
	}

	(void) WORDROOT(WordInfo);

	if (IgnoreCommon) {
	    if (TooCommon(WordInfo)) {
		SawCommon++;
		WordInBlock++;
		continue;
	    } else if (SawCommon) {
		SawCommon = 0;
		WordInfo->WordPlace.Flags |= (WPF_LASTWASCOMMON);
	    }
	}

	if (SawLetters) {
	    SawLetters = 0;
	    WordInfo->WordPlace.Flags |= WPF_LASTHADLETTERS;
	}

	if (SawPunctuation) {
	    SawPunctuation = 0;
	    WordInfo->WordPlace.Flags |= WPF_LASTHADPUNCT;
	}

	/* StuffBefore is the # of chars between the end of the last word and
	 * the start of this one.
	 */
	if (Start > 1L) {
	    if (Start - (LastPos + 1) <= 0) { /* save a byte in the index */
		WordInfo->WordPlace.StuffBefore = 1;
	    } else if (Start - (LastPos + 1) >= 127 ) {
		WordInfo->WordPlace.StuffBefore = 127;
	    } else {
		WordInfo->WordPlace.StuffBefore = Start - (LastPos + 1);
	    }
	} else {
	    WordInfo->WordPlace.StuffBefore = 1; /* i.e., the default */
	}

	WordInfo->WordPlace.WordInBlock = (++WordInBlock);
	WordInfo->WordPlace.BlockInFile = BlockInFile;

	WordInfo->Word[WordInfo->Length] = '\0';

	LastPos = BytesRead - 1;

	ThisOrThat = !ThisOrThat;
	/* toggle between 0 and 1.  Boring life, really */

	break;
    }
    return WordInfo;
}
